from gi.repository import Gtk


class ErrorDialog:
    def __init__(self, parent, text):
        self.dialog = Gtk.MessageDialog(
            transient_for=parent,
            flags=0,
            message_type=Gtk.MessageType.ERROR,
            buttons=Gtk.ButtonsType.OK,
            text=text,
        )

    def run_and_wait(self):
        self.dialog.run()
        self.dialog.destroy()



import caldav

from abeluna.sync.local import LocalServer


class Calendar:
    def __init__(self, uid, name, url, username, password, local_storage):
        self.uid = uid
        self.name = name
        self.url = url
        self.username = username
        self.password = password
        self.local_storage = local_storage

        if self.uid and self.local_storage:
            self.local_server = LocalServer(self.local_storage, self.uid)
        else:
            self.local_server = None

        if self.url:
            self.client = caldav.DAVClient(
                url=self.url,
                username=self.username,
                password=self.password,
            )
            self.calendar = caldav.Calendar(client=self.client, url=self.url)
        else:
            self.client = self.calendar = None

    @property
    def is_local(self):
        return self.client is None

    def validate(self):
        if self.is_local:
            self._validated = True
        else:
            try:
                response = self.client.propfind(self.url)
            except Exception:
                self._validated = False
            else:
                self._validated = response.status in (200, 207)
        return self._validated

    def to_dict(self):
        return {
            'uid': self.uid,
            'name': self.name,
            'url': self.url,
            'username': self.username,
            'password': self.password,
            'local_storage': self.local_storage,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(
            d.get('uid', ''),
            d['name'],
            d['url'],
            d['username'],
            d['password'],
            d.get('local_storage', ''),
        )



import pytz
from gi.repository import GObject, Gdk, Gtk

from abeluna.settings import settings
from abeluna.sync import Calendar, server
from abeluna.widgets import DropdownSelectWidget, ErrorDialog


class CalendarEditor(Gtk.Grid):
    __gsignals__ = {
        'updated-data': (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.data = {}

        self.set_column_spacing(5)
        self.set_row_spacing(5)
        self.set_border_width(10)

        self._create_widgets()

    def set_data(self, data):
        self.data = data
        self.skip_validation_toggle.set_active(False)
        self.is_local_calendar.set_active(False)
        self.server_name_entry.get_buffer().set_text(self.data.get('name', ''), -1)
        self.server_url_entry.get_buffer().set_text(self.data.get('url', ''), -1)
        self.server_username_entry.get_buffer().set_text(self.data.get('username', ''), -1)
        self.server_password_entry.get_buffer().set_text(self.data.get('password', ''), -1)
        if not self.data.get('url'):
            self.is_local_calendar.set_active(True)
            self.update_server_fields()

    def update_server_fields(self):
        is_active = not self.is_local_calendar.get_active()
        for obj in (self.server_url_entry, self.server_username_entry, self.server_password_entry):
            if not is_active:
                obj.get_buffer().set_text('', -1)
            obj.set_sensitive(is_active)

    def get_data(self):
        return self.data

    def _create_widgets(self):
        self.server_name_label = Gtk.Label(label='Calendar Name')
        self.server_name_entry = Gtk.Entry()
        self.attach(self.server_name_label, 0, 0, 2, 1)
        self.attach_next_to(self.server_name_entry, self.server_name_label, Gtk.PositionType.RIGHT, 4, 1)

        self.is_local_calendar_label = Gtk.Label(label='Local Calendar')
        self.is_local_calendar = Gtk.CheckButton()
        self.is_local_calendar.connect('toggled', lambda button: self.update_server_fields())

        self.attach_next_to(self.is_local_calendar_label, self.server_name_label, Gtk.PositionType.BOTTOM, 2, 1)
        self.attach_next_to(self.is_local_calendar, self.is_local_calendar_label, Gtk.PositionType.RIGHT, 4, 1)

        self.server_url_label = Gtk.Label(label='Server URL')
        self.server_url_entry = Gtk.Entry()
        self.attach_next_to(self.server_url_label, self.is_local_calendar_label, Gtk.PositionType.BOTTOM, 2, 1)
        self.attach_next_to(self.server_url_entry, self.server_url_label, Gtk.PositionType.RIGHT, 4, 1)

        self.server_username_label = Gtk.Label(label='Server Username')
        self.server_username_entry = Gtk.Entry()
        self.attach_next_to(self.server_username_label, self.server_url_label, Gtk.PositionType.BOTTOM, 2, 1)
        self.attach_next_to(self.server_username_entry, self.server_username_label, Gtk.PositionType.RIGHT, 4, 1)

        self.server_password_label = Gtk.Label(label='Server Password')
        self.server_password_entry = Gtk.Entry()
        self.server_password_entry.set_visibility(False)
        self.attach_next_to(self.server_password_label, self.server_username_label, Gtk.PositionType.BOTTOM, 2, 1)
        self.attach_next_to(self.server_password_entry, self.server_password_label, Gtk.PositionType.RIGHT, 4, 1)

        self.skip_validation_toggle = Gtk.CheckButton(label='Skip Validation')
        self.attach(self.skip_validation_toggle, 0, 5, 3, 1)

        self.validation_spinner = Gtk.Spinner()
        self.validation_spinner.set_hexpand(False)
        self.validation_spinner.set_vexpand(True)
        self.validation_spinner.set_halign(Gtk.Align.END)
        self.attach_next_to(self.validation_spinner, self.skip_validation_toggle, Gtk.PositionType.RIGHT, 1, 1)

        def save_button_clicked(button):
            self.data['name'] = self.server_name_entry.get_buffer().get_text()
            self.data['url'] = self.server_url_entry.get_buffer().get_text()
            self.data['username'] = self.server_username_entry.get_buffer().get_text()
            self.data['password'] = self.server_password_entry.get_buffer().get_text()
            self.emit('updated-data')

        self.save_button = Gtk.Button('Save!')
        self.save_button.connect('clicked', save_button_clicked)

        self.attach_next_to(self.save_button, self.validation_spinner, Gtk.PositionType.RIGHT, 2, 1)


class SettingsWindow(Gtk.Window):
    def __init__(self, parent, active_child=None):
        Gtk.Window.__init__(self, title='Abeluna Settings')
        self.set_default_size(900, 400)
        self.parent = parent
        self.set_modal(True)
        self.set_transient_for(self.parent)
        self.set_destroy_with_parent(True)
        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.connect('delete-event', lambda obj, event: self.destroy())

        self.main_grid = Gtk.Grid()
        self.main_grid.set_column_spacing(5)
        self.main_grid.set_row_spacing(5)
        self.main_grid.set_border_width(10)
        self.add(self.main_grid)

        self.main_stack = Gtk.Stack()
        self.main_stack_switcher = Gtk.StackSwitcher()
        self.main_stack_switcher.set_stack(self.main_stack)
        self.main_stack_switcher.set_orientation(Gtk.Orientation.VERTICAL)

        self.general_page_grid = Gtk.Grid()
        self.general_page_grid.set_column_homogeneous(True)
        self.general_page_grid.set_column_spacing(5)
        self.general_page_grid.set_row_spacing(5)
        self.general_page_grid.set_vexpand(True)
        self.general_page_grid.set_hexpand(True)

        self.timezone_label = Gtk.Label(label='Timezone')
        timezone_todo_store = Gtk.ListStore(str)
        for tz in pytz.common_timezones:
            timezone_todo_store.append([tz])
        self.timezone_combo = Gtk.ComboBox.new_with_model_and_entry(model=timezone_todo_store)
        self.timezone_combo.set_entry_text_column(0)
        self.timezone_combo.set_id_column(0)

        timezone_completion = Gtk.EntryCompletion()
        timezone_completion.set_model(timezone_todo_store)
        timezone_completion.set_text_column(0)
        timezone_completion.set_minimum_key_length(2)

        def match_timezone_selected(completion, todo_store, it):
            self.timezone_combo.set_active_id(todo_store[it][0])
        timezone_completion.connect('match-selected', match_timezone_selected)

        def match_timezone(completion, string, it):
            to_match = timezone_todo_store[it][0].lower()
            return all(word in to_match for word in string.lower().split())
        timezone_completion.set_match_func(match_timezone)
        self.timezone_combo.get_child().set_completion(timezone_completion)
        self.timezone_combo.set_active_id(settings.TIMEZONE)

        self.general_page_grid.attach(self.timezone_label, 0, 0, 2, 1)
        self.general_page_grid.attach_next_to(self.timezone_combo, self.timezone_label, Gtk.PositionType.RIGHT, 4, 1)

        self.autosync_label = Gtk.Label(label='Autosync Interval')
        self.autosync_selector = DropdownSelectWidget(options=[
            ('-1', 'Never'),
            ('10', 'Every 10 seconds'),
            ('30', 'Every 30 seconds'),
            ('60', 'Every minute'),
            ('600', 'Every 10 minutes'),
            ('1800', 'Every 30 minutes'),
            ('3600', 'Every hour'),
            ('21600', 'Every 6 hours'),
            ('86400', 'Every day'),
            ('604800', 'Every week'),
            ('2419200', 'Every 4 weeks'),
            ('1036800', 'Every year'),
        ])
        self.autosync_selector.set_active_id(settings.AUTOSYNC_INTERVAL)
        self.general_page_grid.attach_next_to(self.autosync_label, self.timezone_label, Gtk.PositionType.BOTTOM, 2, 1)
        self.general_page_grid.attach_next_to(self.autosync_selector, self.autosync_label, Gtk.PositionType.RIGHT, 4, 1)

        self.priority_label = Gtk.Label(label='On conflict')
        self.priority_selector = DropdownSelectWidget(options=[
            ('SERVER', 'Prioritize server'),
            ('CLIENT', 'Prioritize client'),
        ])
        self.priority_selector.set_active_id(settings.PRIORITIZE_ON_CONFLICT)
        self.general_page_grid.attach_next_to(self.priority_label, self.autosync_label, Gtk.PositionType.BOTTOM, 2, 1)
        self.general_page_grid.attach_next_to(self.priority_selector, self.priority_label, Gtk.PositionType.RIGHT, 4, 1)

        self.saved_label = Gtk.Label(label=' ')
        self.saved_label.set_xalign(0.95)
        self.saved_label.set_yalign(0.75)

        def clear_saved_label():
            self.saved_label.set_label(' ')
            return False

        def save_button_clicked(button):
            timezone = self.timezone_combo.get_active_id()
            autosync_interval = self.autosync_selector.get_active_id()
            priority = self.priority_selector.get_active_id()

            failed_settings = []
            for obj, name in (
                (timezone, 'timezone'),
                (autosync_interval, 'autosync interval'),
                (priority, 'priority'),
            ):
                if obj is None:
                    failed_settings.append(name)

            if failed_settings:
                ErrorDialog(
                    self,
                    'Setting for {} is invalid. Please try again.'.format(', '.join(failed_settings)),
                ).run_and_wait()
            else:
                if settings.TIMEZONE != timezone:
                    settings.TIMEZONE = timezone
                    self.parent.todolist_window.rebuild_todolist()
                if settings.AUTOSYNC_INTERVAL != autosync_interval:
                    settings.AUTOSYNC_INTERVAL = autosync_interval
                    server.restart_autosync_thread()
                settings.PRIORITIZE_ON_CONFLICT = priority
                settings.commit()
                self.saved_label.set_label('Saved!')
                GObject.timeout_add_seconds(5, clear_saved_label)

        self.save_button = Gtk.Button('Save!')
        self.save_button.set_margin_top(20)
        self.save_button.connect('clicked', save_button_clicked)

        self.general_page_grid.attach(self.save_button, 4, 4, 2, 1)
        self.general_page_grid.attach_next_to(self.saved_label, self.save_button, Gtk.PositionType.LEFT, 2, 1)

        padding = Gtk.Box()
        padding.set_hexpand(True)
        padding.set_vexpand(True)

        self.general_page_grid.attach_next_to(padding, self.save_button, Gtk.PositionType.TOP, 2, 1)

        self.main_stack.add_titled(self.general_page_grid, 'General', 'General settings')

        self.servers_page_grid = Gtk.Grid()
        self.servers_page_grid.set_column_homogeneous(True)
        self.servers_page_grid.set_column_spacing(5)
        self.servers_page_grid.set_row_spacing(5)

        self.server_todo_store = Gtk.ListStore(str, str, str, str)

        def rebuild_server_todo_store():
            self.server_todo_store.clear()
            for calendar in settings.CALENDARS.values():
                self.server_todo_store.append([
                    calendar['name'],
                    calendar['url'] or 'Local Calendar',
                    calendar['username'] or 'Local Calendar',
                    calendar['uid'],
                ])
        rebuild_server_todo_store()

        self.server_view = Gtk.TreeView(model=self.server_todo_store)
        self.server_view.set_hexpand(True)
        self.server_view.set_headers_clickable(False)
        self.server_view.set_search_column(0)
        for idx, name in enumerate(('Calendar Name', 'URL', 'Username')):
            self.server_view.append_column(Gtk.TreeViewColumn(name, Gtk.CellRendererText(), text=idx))

        self.server_editor_popover = Gtk.Popover()
        self.server_editor = CalendarEditor()

        def server_validation_checker(calendar, data):
            if hasattr(calendar, '_validated'):
                self.server_editor.save_button.set_sensitive(True)
                self.server_editor.validation_spinner.stop()
                if calendar._validated:
                    settings.add_or_update_calendar(data)
                    settings.commit()
                    rebuild_server_todo_store()
                    server.refresh_calendars()
                    self.parent.rebuild_calendarlist()
                    self.server_editor_popover.popdown()
                else:
                    ErrorDialog(
                        self,
                        'Could not connect to server. Please check your settings are correct.',
                    ).run_and_wait()
                return False
            return True

        def on_data_updated(obj):
            data = self.server_editor.get_data()
            if not data['name']:
                ErrorDialog(self, 'Name cannot be empty.').run_and_wait()
                return
            elif (
                not self.server_editor.is_local_calendar.get_active() and
                not all(data[field] for field in ('url', 'username', 'password'))
            ):
                ErrorDialog(self, 'Please enter the URL and credentials for the remote server.').run_and_wait()
                return

            self.server_editor.save_button.set_sensitive(False)
            calendar = Calendar.from_dict(data)
            if self.server_editor.skip_validation_toggle.get_active():
                calendar._validated = True
            else:
                self.server_editor.validation_spinner.start()
                import threading
                threading.Thread(target=calendar.validate).start()
            GObject.timeout_add(100, server_validation_checker, calendar, data)

        self.server_editor.connect('updated-data', on_data_updated)
        self.server_editor_popover.add(self.server_editor)

        def _show_editor_on_row(path):
            uid = self.server_todo_store[path][3]
            self.server_editor.set_data(settings.CALENDARS[uid].copy())

            pos = self.server_view.get_cell_area(path, self.server_view.get_column(0))
            rect = Gdk.Rectangle()
            pos_x, pos_y = self.server_view.convert_bin_window_to_widget_coords(pos.x, pos.y)

            rect.x = 0
            rect.width = self.server_view.get_allocation().width
            rect.y = pos_y
            rect.height = pos.height
            self.server_editor_popover.set_relative_to(self.server_view)
            self.server_editor_popover.set_pointing_to(rect)
            self.server_editor_popover.set_position(Gtk.PositionType.BOTTOM)
            self.server_editor_popover.show_all()
            self.server_editor_popover.popup()

        def on_row_activated(todo_tree_view, path, column):
            _show_editor_on_row(path)

        self.server_view.connect('row-activated', on_row_activated)

        self.scrollable_server_view = Gtk.ScrolledWindow(vexpand=True)
        self.scrollable_server_view.set_propagate_natural_width(True)
        self.scrollable_server_view.set_shadow_type(type=Gtk.ShadowType.ETCHED_OUT)
        self.scrollable_server_view.add(self.server_view)
        self.servers_page_grid.attach(self.scrollable_server_view, 0, 0, 8, 6)

        self.server_add_button = Gtk.Button(label='Add')
        self.server_clone_button = Gtk.Button(label='Clone')
        self.server_edit_button = Gtk.Button(label='Modify')
        self.server_delete_button = Gtk.Button(label='Delete')

        self.server_clone_button.set_sensitive(False)
        self.server_edit_button.set_sensitive(False)
        self.server_delete_button.set_sensitive(False)

        def on_todo_tree_selection_changed(todo_tree_selection):
            current_it = todo_tree_selection.get_selected()[1]
            if current_it is None:
                self.server_clone_button.set_sensitive(False)
                self.server_edit_button.set_sensitive(False)
                self.server_delete_button.set_sensitive(False)
            else:
                self.server_clone_button.set_sensitive(True)
                self.server_edit_button.set_sensitive(True)
                self.server_delete_button.set_sensitive(True)
        self.server_view.get_selection().connect('changed', on_todo_tree_selection_changed)

        def on_server_create(button, clone_selected):
            if clone_selected:
                it = self.server_view.get_selection().get_selected()[1]
                if it is None:
                    return
                uid = self.server_todo_store[it][3]
                data = settings.CALENDARS[uid].copy()
                data.pop('uid')
            else:
                data = {}

            self.server_editor.set_data(data)
            rect = Gdk.Rectangle()
            rect.width = button.get_allocation().width
            rect.height = button.get_allocation().height

            self.server_editor_popover.set_relative_to(button)
            self.server_editor_popover.set_pointing_to(rect)
            self.server_editor_popover.set_position(Gtk.PositionType.TOP)
            self.server_editor_popover.show_all()
            self.server_editor_popover.popup()
        self.server_add_button.connect('clicked', on_server_create, False)
        self.server_clone_button.connect('clicked', on_server_create, True)

        def on_server_edit(button):
            it = self.server_view.get_selection().get_selected()[1]
            if it is None:
                return
            _show_editor_on_row(self.server_todo_store.get_path(it))
        self.server_edit_button.connect('clicked', on_server_edit)

        def on_server_delete(button):
            it = self.server_view.get_selection().get_selected()[1]
            if it is None:
                return
            settings.delete_calendar(self.server_todo_store[it][3])
            settings.commit()
            rebuild_server_todo_store()
            server.refresh_calendars()
            self.parent.rebuild_calendarlist()

        self.server_delete_button.connect('clicked', on_server_delete)

        self.servers_page_grid.attach_next_to(
            self.server_add_button, self.scrollable_server_view, Gtk.PositionType.BOTTOM, 2, 1,
        )
        self.servers_page_grid.attach_next_to(
            self.server_clone_button, self.server_add_button, Gtk.PositionType.RIGHT, 2, 1,
        )
        self.servers_page_grid.attach_next_to(
            self.server_edit_button, self.server_clone_button, Gtk.PositionType.RIGHT, 2, 1,
        )
        self.servers_page_grid.attach_next_to(
            self.server_delete_button, self.server_edit_button, Gtk.PositionType.RIGHT, 2, 1,
        )

        self.main_stack.add_titled(self.servers_page_grid, 'Calendars', 'Calendar settings')

        def on_close_window(button):
            self.destroy()

        self.close_button = Gtk.Button(label='Close')
        self.close_button.connect('clicked', on_close_window)

        padding = Gtk.Box()
        padding.set_vexpand(True)
        self.main_grid.attach(self.main_stack_switcher, 0, 0, 3, 2)
        self.main_grid.attach_next_to(padding, self.main_stack_switcher, Gtk.PositionType.BOTTOM, 3, 1)
        self.main_grid.attach_next_to(self.close_button, padding, Gtk.PositionType.BOTTOM, 3, 1)
        self.main_grid.attach(self.main_stack, 3, 0, 5, 4)

        self.show_all()

        if active_child is not None:
            self.main_stack.set_visible_child_name(active_child)



import os
import sqlite3
from collections import namedtuple
from functools import partial

import icalendar


class LocalTodo(namedtuple('LocalTodo', 'uid local_vtodo remote_vtodo')):
    def __eq__(self, other):
        if isinstance(other, str):
            return self.uid == other
        return self.uid == other.uid

    def __hash__(self):
        return self.uid


class LocalServer:
    def __init__(self, path, calendar_name):
        self.path = path
        self.calendar = calendar_name
        self.conn = partial(sqlite3.connect, os.path.join(self.path, '{}.db'.format(self.calendar)))

        with self.conn() as c:
            c.execute('''
                CREATE TABLE IF NOT EXISTS todo
                (uid TEXT PRIMARY KEY, local_vtodo TEXT, remote_vtodo TEXT)
            ''')
            c.commit()

        self.todolist = []

    def _sanitize_uid(self, uid):
        return str(uid)

    def todos(self, include_deleted=False):
        def create_ical(val):
            if val is None:
                return val
            return icalendar.Calendar.from_ical(val)

        with self.conn() as c:
            data = c.execute(
                '''
                SELECT * FROM todo
                {where}
                '''.format(where='' if include_deleted else 'WHERE local_vtodo IS NOT NULL'),
            ).fetchall()

        return [
            LocalTodo(
                uid=item[0],
                local_vtodo=create_ical(item[1]),
                remote_vtodo=create_ical(item[2]),
            ) for item in data
        ]

    def update_todo_from_server(self, vtodo):
        ical = vtodo.to_ical().decode()
        uid = self._sanitize_uid(vtodo['UID'])
        with self.conn() as c:
            c.execute(
                '''
                INSERT OR IGNORE INTO todo
                VALUES (?, NULL, NULL)
                ''',
                (uid,),
            )
            c.execute(
                '''
                UPDATE todo
                SET local_vtodo=?, remote_vtodo=?
                WHERE uid=?
                ''',
                (ical, ical, uid),
            )
            c.commit()

    def update_todo_from_client(self, vtodo):  # also includes creating the todo
        ical = vtodo.to_ical().decode()
        uid = self._sanitize_uid(vtodo['UID'])
        with self.conn() as c:
            c.execute(
                '''
                INSERT OR IGNORE INTO todo
                VALUES (?, NULL, NULL)
                ''',
                (uid,),
            )
            c.execute(
                '''
                UPDATE todo
                SET local_vtodo=?
                WHERE uid=?
                ''',
                (ical, uid),
            )
            c.commit()

    def delete_todo_from_server(self, vtodo):
        with self.conn() as c:
            c.execute(
                '''
                DELETE FROM todo
                WHERE uid = ?
                ''',
                (self._sanitize_uid(vtodo['UID']),),
            )
            c.commit()

    def delete_todo_from_client(self, vtodo):
        uid = self._sanitize_uid(vtodo['UID'])
        with self.conn() as c:
            c.execute(
                '''
                DELETE FROM todo
                WHERE uid = ? and remote_vtodo IS NULL
                ''',
                (uid,),
            )
            c.execute(
                '''
                UPDATE todo
                SET local_vtodo = NULL
                WHERE uid = ? and remote_vtodo IS NOT NULL
                ''',
                (uid,),
            )
            c.commit()



from setuptools import find_packages, setup

with open('README.md') as f:
    readme = f.read()

setup(
    name='abeluna',
    version='1.0.4',
    entry_points={
        'gui_scripts': [
            'abeluna = abeluna.main:main',
        ],
    },
    author='Evan Zhang',
    install_requires=['pygobject', 'humanize', 'icalendar', 'caldav'],
    include_package_data=True,
    description='A simple GUI to-do/task manager with CalDAV support.',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/Ninjaclasher/abeluna',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: X11 Applications :: GTK',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



from gi.repository import Gtk


class DropdownSelectWidget(Gtk.ComboBox):
    def __init__(self, *args, **kwargs):
        options = kwargs.pop('options')

        if len(options) < 1:
            raise ValueError('no options')

        self.id_index = kwargs.pop('id_index', 0)
        self.value_index = kwargs.pop('value_index', 1)

        if not isinstance(options[0][self.id_index], str):
            raise ValueError('id must be of type string')

        self.store = Gtk.ListStore(*[type(idx) for idx in options[0]])
        for option in options:
            self.store.append(option)
        kwargs['model'] = self.store

        super().__init__(*args, **kwargs)

        renderer_text = Gtk.CellRendererText()
        self.pack_start(renderer_text, True)
        self.add_attribute(renderer_text, 'text', self.value_index)

        self.set_id_column(self.id_index)



import datetime

import icalendar
import pytz

from abeluna.settings import settings


def generate_vtimezone(timezone=None, for_date=None):
    # In case the setting value changes.
    if timezone is None:
        timezone = settings.TIMEZONE

    def _vtimezone_without_dst(std, timezone):
        vtimezone = icalendar.Timezone(tzid=timezone)
        standard = icalendar.TimezoneStandard()
        utc_offset, dst_offset, tz_name = std[1]
        standard.add('dtstart', std[0])
        standard.add('tzoffsetfrom', utc_offset)
        standard.add('tzoffsetto', utc_offset)
        standard.add('tzname', tz_name)
        vtimezone.add_component(standard)
        return vtimezone

    def _vtimezone_with_dst(dst1, std1, dst2, std2, timezone):
        vtimezone = icalendar.Timezone(tzid=timezone)
        daylight = icalendar.TimezoneDaylight()
        utc_offset, dst_offset, tz_name = dst1[1]
        offsetfrom = std1[1][0]
        daylight.add('dtstart', dst1[0] + offsetfrom)
        daylight.add('rdate', dst1[0] + offsetfrom)
        daylight.add('rdate', dst2[0] + offsetfrom)
        daylight.add('tzoffsetfrom', offsetfrom)
        daylight.add('tzoffsetto', utc_offset)
        daylight.add('tzname', tz_name)
        vtimezone.add_component(daylight)

        standard = icalendar.TimezoneStandard()
        utc_offset, dst_offset, tz_name = std1[1]
        offsetfrom = dst1[1][0]
        standard.add('dtstart', std1[0] + offsetfrom)
        standard.add('rdate', std1[0] + offsetfrom)
        standard.add('rdate', std2[0] + offsetfrom)
        standard.add('tzoffsetfrom', offsetfrom)
        standard.add('tzoffsetto', utc_offset)
        standard.add('tzname', tz_name)
        vtimezone.add_component(standard)
        return vtimezone

    if not timezone or 'utc' in timezone.lower():
        return None
    if for_date is None:
        for_date = datetime.datetime.now()
    z = pytz.timezone(timezone)
    transitions = zip(z._utc_transition_times, z._transition_info)
    try:
        dst1, std1, dst2, std2 = filter(lambda x: x[0].year in (for_date.year, for_date.year + 1),
                                        transitions)
        if dst1[1][1].seconds == 0:
            return _vtimezone_with_dst(std1, dst1, std2, dst2, timezone)
        else:
            return _vtimezone_with_dst(dst1, std1, dst2, std2, timezone)

    except Exception:
        std = transitions[-1]
        if std[0].year > for_date.year:
            return None
        return _vtimezone_without_dst(std, timezone)


def colour_text(text, colour):
    return '<span foreground="{colour}">{text}</span>'.format(colour=colour, text=text)



import datetime
import queue
import threading
import time
import uuid
from collections import defaultdict
from functools import wraps

import caldav
import icalendar

from abeluna.settings import settings
from abeluna.sync.calendar import Calendar
from abeluna.util import generate_vtimezone


class Task:
    def __init__(self, progress=0):
        self.uid = uuid.uuid4().hex
        self._progress = progress

    def __str__(self):
        return self.uid

    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, value):
        self._progress = value

    @property
    def completed(self):
        return self._progress == 100

    @completed.setter
    def completed(self, value):
        self._progress = 100 if value else 0


def background_task(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        try:
            task = kwargs.pop('task', Task())
            delay_abs = kwargs.pop('delay', -1) + self.timefunc()
            self.task_queue.put_nowait((delay_abs, method, task, (self, task) + args, kwargs))
        except queue.Full:
            return None
        else:
            return task
    return wrapper


class SynchronizationServer:
    def __init__(self):
        self.task_queue = queue.PriorityQueue()

        self.todolist = defaultdict(list)

        self._stop_lock = threading.RLock()

        self._worker_stop = threading.Event()
        self._worker_thread = threading.Thread(target=self.worker_run)
        self._worker_thread.start()
        self._general_lock = threading.RLock()
        self._update_todo_skip = {}
        self.timefunc = time.time

        self._autosync_stop = threading.Event()
        self._autosync_thread = None
        self.restart_autosync_thread()
        self.last_sync = None
        self._sync_lock = threading.RLock()
        self._sync_callbacks = []

        self.calendars = {}

    def stop_all(self):
        with self._stop_lock:
            self._worker_stop.set()
            self._autosync_stop.set()
        self._worker_thread.join()
        self._autosync_thread.join()

    def worker_run(self):
        timeout = 0
        while not self._worker_stop.wait(timeout=timeout):
            try:
                delay_abs, method, task, args, kwargs = self.task_queue.get_nowait()
                if delay_abs > self.timefunc():
                    self.task_queue.put_nowait((delay_abs, method, task, args, kwargs))
                    raise queue.Empty()
            except queue.Empty:
                timeout = 0.25
            else:
                timeout = 0
                try:
                    task.completed = method(*args, **kwargs)
                except KeyboardInterrupt:
                    break
                except Exception:  # catch all
                    task.completed = False
                    import traceback
                    traceback.print_exc()

                # print('Process background task:', method, task)
                self.task_queue.task_done()

    def restart_autosync_thread(self):
        with self._stop_lock:
            if self._autosync_stop.is_set():
                return
            self._autosync_stop.set()
            if self._autosync_thread is not None:
                self._autosync_thread.join()
            self._autosync_stop.clear()

            self._autosync_thread = threading.Thread(target=self.autosync_run)
            self._autosync_thread.start()

    def autosync_run(self):
        if settings.AUTOSYNC_INTERVAL == '-1':
            return
        while not self._autosync_stop.wait(timeout=int(settings.AUTOSYNC_INTERVAL)):
            try:
                self._synchronize_todolist()
            except KeyboardInterrupt:
                break
            except Exception:  # catch all
                import traceback
                traceback.print_exc()

    def sync_connect(self, callback, *args, **kwargs):
        self._sync_callbacks.append((callback, args, kwargs))

    def _merge_todo(self, uid, local_copy_of_local, local_copy_of_remote, remote_copy_of_remote):
        # Nothing changed, don't touch anything.
        if local_copy_of_local.to_ical() == remote_copy_of_remote.to_ical():
            return False, remote_copy_of_remote
        # Nothing changed server side, so use the client todo if there are any updates.
        elif local_copy_of_remote.to_ical() == remote_copy_of_remote.to_ical():
            # print(uid, 'was changed locally but not changed on remote. Pushing to remote...')
            return True, local_copy_of_local
        # Something changed server side and client side, so we will prioritize the server.
        elif local_copy_of_remote.to_ical() != remote_copy_of_remote.to_ical():
            # print(uid, 'was changed both locally and on remote. Merging...')

            # User prioritizes server.
            for key, value in local_copy_of_local.items():
                try:
                    remote_copy_of_remote_value = remote_copy_of_remote[key].to_ical()
                except KeyError:
                    remote_copy_of_remote_value = None
                try:
                    local_copy_of_remote_value = local_copy_of_remote[key].to_ical()
                except KeyError:
                    local_copy_of_remote_value = None
                try:
                    local_copy_of_local_value = local_copy_of_local[key].to_ical()
                except KeyError:
                    local_copy_of_local_value = None

                # Nothing changed server side, so use client value.
                if (
                    settings.PRIORITIZE_ON_CONFLICT == 'SERVER' and
                    local_copy_of_remote_value == remote_copy_of_remote_value
                ):
                    remote_copy_of_remote[key] = value
                # Something changed client side, so use client value.
                elif (
                    settings.PRIORITIZE_ON_CONFLICT == 'CLIENT' and
                    local_copy_of_remote_value != local_copy_of_local_value
                ):
                    remote_copy_of_remote[key] = value
            return True, remote_copy_of_remote
        # How did we get here...
        else:
            assert False

    def _synchronize_todolist(self):
        with self._sync_lock:
            for cb, args, kwargs in self._sync_callbacks:
                cb(*args, mode='PRE_SYNC', **kwargs)

            for cal in self.calendars.values():
                try:
                    if cal.is_local:
                        continue
                    remote_todos = cal.calendar.todos(include_completed=True)
                    local_todos = cal.local_server.todos(include_deleted=True)

                    remote_uids = set()
                    for remote_todo in remote_todos:
                        remote_ical = remote_todo.icalendar_instance
                        new_cal = remote_ical.copy()
                        has_todo_component = False
                        updated_todo_component = False
                        for remote_item in remote_todo.icalendar_instance.subcomponents:
                            # Keep all non-todo items unconditionally in case there are any.
                            if not isinstance(remote_item, icalendar.Todo):
                                new_cal.add_component(remote_item)
                                continue

                            uid = str(remote_item['UID'])
                            remote_uids.add(uid)
                            try:
                                local_item = local_todos[local_todos.index(uid)]
                            except ValueError:
                                # print(uid, 'does not exist locally. Creating...')
                                # Item exists on the server but does not exist locally AND was not deleted locally.
                                has_todo_component = True
                                new_cal.add_component(remote_item)
                                cal.local_server.update_todo_from_server(remote_item)
                            else:
                                # Item exists on the server but does not exist locally AND was deleted locally.
                                if local_item.local_vtodo is None:
                                    # print(uid, 'was deleted locally. Deleting from server...')
                                    updated_todo_component = True
                                    cal.local_server.delete_todo_from_server(remote_item)
                                # Item exists on both the server and the client, compare the todos
                                else:
                                    has_todo_component = True
                                    updated, item_to_use = self._merge_todo(
                                        uid,
                                        local_item.local_vtodo,
                                        local_item.remote_vtodo,
                                        remote_item,
                                    )
                                    updated_todo_component |= updated

                                    new_cal.add_component(item_to_use)
                                    cal.local_server.update_todo_from_server(item_to_use)

                        if not has_todo_component:
                            remote_todo.delete()
                        elif updated_todo_component:
                            remote_todo.icalendar_instance = new_cal
                            remote_todo.save()

                    for local_item in local_todos:
                        # Item existed on server, so it was already processed.
                        if local_item.uid in remote_uids:
                            continue

                        # Item has a record of being on the server, but it doesn't exist on the server anymore.
                        # We can only assume it was deleted server-side.
                        if local_item.remote_vtodo is not None:
                            # print(local_item.uid, 'was deleted on remote. Deleting locally...')
                            cal.local_server.delete_todo_from_server(local_item.remote_vtodo)
                        # Item exists on client, has never existed on server, so create and push to the server.
                        else:
                            # print(local_item.uid, 'was created locally. Pushing to remote...')
                            vcal = icalendar.Calendar()
                            vcal.add('VERSION', '2.0')
                            vcal.add('PRODID', '-//Abeluna//NONSGML v1.0//EN')
                            vcal.add('CALSCALE', 'GREGORIAN')
                            vcal.add_component(generate_vtimezone())
                            vcal.add_component(local_item.local_vtodo)
                            caldav.Todo(cal.client, data=vcal, parent=cal.calendar, id=local_item.uid).save()
                            cal.local_server.update_todo_from_server(local_item.local_vtodo)
                except Exception:
                    import traceback
                    traceback.print_exc()

            self.initialize_todolist()
            self.last_sync = datetime.datetime.now()
            for cb, args, kwargs in self._sync_callbacks:
                cb(*args, mode='POST_SYNC', **kwargs)

    @background_task
    def synchronize_todolist(self, task):
        self._synchronize_todolist()
        return True

    def initialize_todolist(self, uid=None):
        with self._sync_lock:
            if uid is not None:
                try:
                    calendar = self.calendars[uid]
                except KeyError:
                    pass
                else:
                    self.todolist[uid] = [item.local_vtodo for item in calendar.local_server.todos()]
            else:
                self.todolist.clear()
                for uid, cal in self.calendars.items():
                    self.todolist[uid] = [item.local_vtodo for item in cal.local_server.todos()]

    def refresh_calendars(self):
        new_calendars = {}
        for uid, cal_dict in settings.CALENDARS.items():
            cal = Calendar.from_dict(cal_dict)
            new_calendars[uid] = cal

        with self._sync_lock:
            self.calendars = new_calendars
            self.initialize_todolist()

    @background_task
    def update_todo(self, task, vtodo, cal_uid, postpone=True):
        uid = str(vtodo['UID'])
        _time = self.timefunc()
        with self._general_lock:
            # Some sketchy logic to "batch" saves together.
            # E.g when the user is editting a textbox, don't save after every keystroke.
            if postpone:
                if uid not in self._update_todo_skip:
                    self.update_todo(vtodo, cal_uid, task=task, postpone=False)
                self._update_todo_skip[uid] = _time
                return False
            else:
                time_diff = int(settings.SAVE_INTERVAL) - (_time - self._update_todo_skip.get(uid, 0))
                if time_diff > 0:
                    self.update_todo(vtodo, cal_uid, task=task, postpone=False, delay=time_diff)
                    return False
                else:
                    try:
                        self._update_todo_skip.pop(uid)
                    except KeyError:
                        pass

        with self._sync_lock:
            self.calendars[cal_uid].local_server.update_todo_from_client(vtodo)
            self.initialize_todolist(uid=cal_uid)
        return True

    @background_task
    def delete_todo(self, task, vtodo, cal_uid):
        with self._sync_lock:
            self.calendars[cal_uid].local_server.delete_todo_from_client(vtodo)
            self.initialize_todolist(uid=cal_uid)
        return True


server = SynchronizationServer()
